/*! @file rx_pick.h
	
	@brief OpenGL̃ZNV[hg_sbN
		   `IɂΉ
 
	@author Makoto Fujisawa
	@date   2008-06, 2010-03, 2013-02
*/

#pragma warning (disable: 4244)

#ifndef _RX_PICK_H_
#define _RX_PICK_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cstdio>
#include <cstdlib>

#include <vector>
#include <algorithm>

#include <GL/glut.h>

#include "rx_utility.h"


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
using namespace std;

//! sbNꂽIuWFNgɊւ
struct rxPickInfo
{
	GLuint name;		//!< qbgIuWFNg̖O
	float min_depth;	//!< v~eBũfvXl̍ŏl
	float max_depth;	//!< v~eBũfvXl̍ől
};

static inline bool CompFuncPickInfo(rxPickInfo a, rxPickInfo b)
{
	return a.min_depth < b.min_depth;
}

//-----------------------------------------------------------------------------
// OpenGLɂIuWFNgsbN
//-----------------------------------------------------------------------------
class rxGLPick
{
	void (*m_fpDraw)(void*);	//!< sbNp`֐
	void (*m_fpProj)(void*);	//!< eϊ֐
	void *m_pDrawFuncPtr;		//!< `֐Ăяo߂̃|C^(o֐|C^p)
	void *m_pProjFuncPtr;		//!< eϊ֐Ăяo߂̃|C^(o֐|C^p)
 
	int m_iSelBufferSize;		//!< ZNVobt@̃TCY
	GLuint* m_pSelBuffer;		//!< ZNVobt@

	int m_iLastPick;			//!< ŌɃsbNꂽIuWFNg̔ԍ
 
public:
	rxGLPick()
	{
		m_iSelBufferSize = 4096;
		m_pSelBuffer = new GLuint[m_iSelBufferSize];
	}
	~rxGLPick()
	{
		if(m_pSelBuffer) delete [] m_pSelBuffer;
	}
 
	void Set(void (*draw)(void*), void* draw_ptr, void (*proj)(void*), void* proj_ptr)
	{
		m_fpDraw = draw;
		m_fpProj = proj;
		m_pDrawFuncPtr = draw_ptr;
		m_pProjFuncPtr = proj_ptr;
	}
 
protected:
	/*!
	 * OpenGLɂsbNŃqbĝŏ̃fvXl̂I
	 * @param hits qbg
	 * @param buf  Iobt@
	 * @return qbgIuWFNgԍ
	 */
	vector<rxPickInfo> selectHits(GLint nhits, GLuint buf[])
	{
		vector<rxPickInfo> hits;
 
		float depth_min = 100.0f;
		float depth1 = 1.0f;
		float depth2 = 1.0f;
 
		GLuint depth_name;
		GLuint *ptr;
		
		// qbgf[^Ȃ
		if(nhits <= 0) return hits;
		
		hits.resize(nhits);

		// |C^Ɨpptr֓nD
		ptr = (GLuint*)buf;
		for(int i = 0; i < nhits; ++i){
			// qbgIuWFNg̖O
			depth_name = *ptr;
			ptr++;

			// qbgv~eBũfvXl̍ŏl
			depth1 = (float) *ptr/0x7fffffff;
			ptr++;

			// qbgv~eBũfvXl̍ő
			depth2 = (float) *ptr/0x7fffffff;
			ptr++;

			hits[i].name = (int)(*ptr);
			hits[i].min_depth = depth1;
			hits[i].max_depth = depth2;

			ptr++;
		}

		return hits;
	}
 
	/*!
	 * }EXI
	 * @param[in] x,y I𒆐SW(}EXWn)
	 * @param[in] w,h I͈
	 * @retval int sbNIuWFNgԍ
	 */
	vector<rxPickInfo> pick(int x, int y, int w, int h)
	{
		GLint viewport[4];
 
		glGetIntegerv(GL_VIEWPORT, viewport);
		glSelectBuffer(m_iSelBufferSize, m_pSelBuffer);
 
		glRenderMode(GL_SELECT);
 
		glInitNames();
		glPushName(0);
 
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
 
		gluPickMatrix(x, viewport[3]-y, w, h, viewport);
		m_fpProj(m_pProjFuncPtr);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
 
		m_fpDraw(m_pDrawFuncPtr);
		glLoadName(0);
 
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glPopName();
 
		GLint nhits;
		nhits = glRenderMode(GL_RENDER);
		
		vector<rxPickInfo> hits = selectHits(nhits, m_pSelBuffer);
		glMatrixMode(GL_MODELVIEW);

		if(nhits > 0) m_iLastPick = hits[0].name;
		return hits;
	}
 
 
public:
	/*!
	 * IuWFNg̃}EXI
	 * @param[in] x,y }EXW
	 * @retval true sbN
	 * @retval false sbNs
	 */
	int Pick(int x, int y)
	{
		vector<rxPickInfo> hits = pick(x, y, 16, 16);
		if(hits.empty()){
			m_iLastPick = -1;
			return -1;
		}
		else{
			std::sort(hits.begin(), hits.end(), CompFuncPickInfo);
			m_iLastPick = hits[0].name;
			return hits[0].name;
		}
 
		return -1;
	}

	/*!
	 * IuWFNg̃}EXI(`͈͎w, I)
	 * @param[in] sx,sy n_
	 * @param[in] ex,ey I_
	 * @retval true sbN
	 * @retval false sbNs
	 */
	vector<rxPickInfo> Pick(int sx, int sy, int ex, int ey)
	{
		int x, y, w, h;

		if(ex < sx) RX_SWAP(ex, sx);
		if(ey < sy) RX_SWAP(ey, sy);

		x = (ex+sx)/2;
		y = (ey+sy)/2;
		w = ex-sx;
		h = ey-sy;

		return pick(x, y, w, h);
	}

	/*!
	 * ŌɃsbNꂽIuWFNg̔ԍԂ
	 * @return sbNꂽIuWFNg̔ԍ
	 */
	int GetLastPick(void) const
	{
		return m_iLastPick;
	}
};


/*!
 * }EXhbÖ̕`(`ƒ)
 * @param[in] type 1Œ,2ŋ`
 * @param[in] p0,p1 }EXhbOJn_CI_
 */
static inline void DrawRubber(int type, Vec2 p0, Vec2 p1, int w, int h)
{
	glDisable(GL_LIGHTING);
	glColor3f(0.0, 0.0, 0.0);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, w, h, 0);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	if(type == 1){		// `
		glColor3d(1.0, 1.0, 1.0);
		glPushMatrix();
		glBegin(GL_LINE_LOOP);
		glVertex2d(p0[0], p0[1]);
		glVertex2d(p1[0], p0[1]);
		glVertex2d(p1[0], p1[1]);
		glVertex2d(p0[0], p1[1]);
		glEnd();
		glPopMatrix();
	}
	else if(type == 2){	// 
		glDisable(GL_LIGHTING);
		glColor3d(0.0, 0.0, 0.0);
		glBegin(GL_LINES);
		glVertex2dv(p0.data);
		glVertex2dv(p1.data);
		glEnd();
	}
	
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

}



#endif // #ifndef _RX_PICK_H_