/*! 
  @file rx_png.h
	
  @brief PNGt@Cǂݍ݁Co
	- libpng : http://www.libpng.org/pub/png/libpng.html
	- zlib : http://www.zlib.net/
 
  @author Makoto Fujisawa
  @date 2011-06
*/
// FILE --rx_png.h--

#ifndef _RX_PNG_H_
#define _RX_PNG_H_



//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cstdio>
#include <string>

#include <png.h>


//-----------------------------------------------------------------------------
// PNGt@C̓ǂݍ݂Ə
//-----------------------------------------------------------------------------
inline void ReadPngCallback(png_structp png_ptr, png_uint_32 row_number, int pass)
{
	
}

/*!
 * PNGt@C̓ǂݍ
 * @param[in] fn t@C
 * @param[out] w,h 摜TCY
 * @param[out] c 摜̐F[x
 * @return WJς݉摜f[^
 */
static unsigned char* ReadPngFile(const std::string &fn, int &w, int &h, int &c)
{
	png_FILE_p fp;
	unsigned char* img = 0;	// o̓f[^

	// t@CoCi[hŊJ
	if((fp = fopen(fn.c_str(), "rb")) == NULL){
		fprintf(stderr, "png error : cannot open %s file\n", fn.c_str());
		return 0;
	}

	png_structp read_ptr;
	png_infop read_info_ptr;

	// PNGǂݍ݃IuWFNg
	read_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	read_info_ptr = png_create_info_struct(read_ptr);
	
	// ǂݍރf[^o^
	png_init_io(read_ptr, fp);

	// ǂݍݎ̃R[obN̐ݒ(ǂݍݏ󋵂\ꍇȂ)
	png_set_read_status_fn(read_ptr, NULL);

	// wb_̓ǂݍ
	png_read_info(read_ptr, read_info_ptr);

	// 摜̎擾
	png_uint_32 width, height;
	int bit_depth, color_type;
	int interlace_type, compression_type, filter_type;

	if(png_get_IHDR(read_ptr, read_info_ptr, &width, &height, &bit_depth,
					&color_type, &interlace_type, &compression_type, &filter_type))
	{
		w = (int)width;
		h = (int)height;
		c = 0;

		// `l
		c += (color_type & PNG_COLOR_MASK_COLOR) ? 3 : 0;
		c += (color_type & PNG_COLOR_MASK_ALPHA) ? 1 : 0;
		
		if(c && (color_type & PNG_COLOR_MASK_PALETTE)){
			png_set_palette_to_rgb(read_ptr);
			c = 4;
		}
		else if(!c){	// color_type == PNG_COLOR_TYPE_GRAY
			c = 1;
		}

		if(!c){
			png_destroy_read_struct(&read_ptr, &read_info_ptr, NULL);
			fclose(fp);
			return 0;
		}

		// esւ̃|C^gĉ摜ǂݍ
		img = new unsigned char[w*h*c];
		unsigned char **lines = new unsigned char*[h];
		for(int j = 0; j < h; ++j) lines[j] = &img[j*w*c];

		png_read_image(read_ptr, lines);

		delete [] lines;
	}
	
	png_destroy_read_struct(&read_ptr, &read_info_ptr, NULL);
	fclose(fp);

	return img;
}

/*!
 * PNGt@C̏
 * @param[in] fn t@C
 * @param[in] img 摜f[^
 * @param[in] w,h 摜TCY
 * @param[in] c 摜̐F[x
 * @param[in] quality ki[0,100]
 */
static int WritePngFile(const std::string &fn, unsigned char *img, int w, int h, int c)
{
	png_FILE_p fp;

	// t@CoCi[hŊJ
	if((fp = fopen(fn.c_str(), "wb")) == NULL){
		fprintf(stderr, "png error : cannot open %s file\n", fn.c_str());
		return 0;
	}

	png_structp write_ptr;
	png_infop write_info_ptr;

	// PNGǂݍ݃IuWFNg
	write_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	write_info_ptr = png_create_info_struct(write_ptr);

	if(setjmp(png_jmpbuf(write_ptr))){
		fclose(fp);
		return 0;
	}
	
	// ރt@Co^
	png_init_io(write_ptr, fp);

	// ǂݍݎ̃R[obN̐ݒ(ǂݍݏ󋵂\ꍇȂ)
	png_set_write_status_fn(write_ptr, NULL);

	// J[^Cv
	int color_type = 0;
	if(c == 1) color_type = PNG_COLOR_TYPE_GRAY;
	else if(c == 3) color_type = PNG_COLOR_TYPE_RGB;
	else if(c == 4) color_type = PNG_COLOR_TYPE_RGB_ALPHA;
	
	// wb_̓o^
	png_set_IHDR(write_ptr, write_info_ptr, w, h, 8, color_type, 
				 PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_BASE);

	png_color_8 sig_bit;
	sig_bit.red = 8;
	sig_bit.green = 8;
	sig_bit.blue = 8;
	sig_bit.alpha = 0;
	png_set_sBIT(write_ptr, write_info_ptr, &sig_bit);

	// PNGɏ܂Rg
	png_text text_ptr[1];
	text_ptr[0].key = "Description";
	text_ptr[0].text = "Saved by libpng";
	text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
	png_set_text(write_ptr, write_info_ptr, text_ptr, 1);

	// wb_̏
	png_write_info(write_ptr, write_info_ptr);

	//png_set_bgr(write_ptr);

	// esւ̃|C^gĉ摜ǂݍ
	unsigned char **lines = new unsigned char*[h];
	for(int j = 0; j < h; ++j) lines[j] = &img[j*w*c];

	png_write_image(write_ptr, lines);

	delete [] lines;

	png_write_end(write_ptr, write_info_ptr);
	png_destroy_write_struct(&write_ptr, &write_info_ptr);
	fclose(fp);

	return 1;
}




#endif // #ifdef _RX_PNG_H_
