/*!
  @file rx_mc.h
	
  @brief A֐\ʂ̃|S(MC@)
	
	http://local.wasp.uwa.edu.au/~pbourke/geometry/polygonise/
 
  @author Raghavendra Chandrashekara (basesd on source code
			provided by Paul Bourke and Cory Gene Bloyd)
  @date   2010-03
*/


#ifndef _RX_MC_MESH_H_
#define _RX_MC_MESH_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cstdlib>

// OpenGL
#include <GL/glew.h>
#include <GL/glut.h>

// STL
#include <map>
#include <vector>
#include <string>

#include <iostream>

#include "rx_utility.h"
#include "rx_mesh.h"

#include "rx_cu_common.cuh"


//-----------------------------------------------------------------------------
// O
//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
typedef unsigned int uint;

#ifndef RXREAL
	#define RXREAL float
#endif


struct RxVertexID
{
	uint newID;
	double x, y, z;
};

typedef std::map<uint, RxVertexID> ID2VertexID;

struct RxTriangle
{
	uint pointID[3];
};

typedef std::vector<RxTriangle> RxTriangleVector;

struct RxScalarField
{
	uint iNum[3];
	Vec3 fWidth;
	Vec3 fMin;
};


//-----------------------------------------------------------------------------
// rxMCMeshCPUNX
//-----------------------------------------------------------------------------
class rxMCMeshCPU
{
public:
	// RXgN^
	rxMCMeshCPU();

	// fXgN^
	~rxMCMeshCPU();
	
	//! A֐Op`bV𐶐
	bool CreateMesh(RXREAL (*func)(double, double, double, void*), void* func_ptr, Vec3 min_p, double h, int n[3], RXREAL threshold, 
					vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Op`bV𐶐
	bool CreateMeshV(RXREAL *field, Vec3 min_p, double h, int n[3], RXREAL threshold, 
					 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[瓙lʃbV
	void GenerateSurface(const RxScalarField sf, RXREAL *field, RXREAL threshold, 
						 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! ֐Tv{[쐬ēlʃbV
	void GenerateSurfaceV(const RxScalarField sf, RXREAL (*func)(double, double, double, void*), void* func_ptr, RXREAL threshold, 
						  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! ֐瓙lʃbV
	void GenerateSurfaceF(const RxScalarField sf, RXREAL (*func)(double, double, double, void*), void* func_ptr, RXREAL threshold, 
						  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! lʍ쐬trueԂ
	bool IsSurfaceValid() const { return m_bValidSurface; }

	//! 쐬lʃbV̔j
	void DeleteSurface();

	//! bVɗpObh̑傫(bV쐬ĂȂꍇ͕Ԓl-1)
	int GetVolumeLengths(double& fVolLengthX, double& fVolLengthY, double& fVolLengthZ);

	// 쐬bV̏
	uint GetNumVertices(void) const { return m_nVertices; }
	uint GetNumTriangles(void) const { return m_nTriangles; }
	uint GetNumNormals(void) const { return m_nNormals; }

protected:
	// MARK:oϐ
	uint m_nVertices;	//!< lʃbV̒_
	uint m_nNormals;	//!< lʃbV̒_@(쐬Ă @=_)
	uint m_nTriangles;	//!< lʃbV̎Op`|S

	ID2VertexID m_i2pt3idVertices;			//!< lʂ`钸_̃Xg
	RxTriangleVector m_trivecTriangles;		//!< Op`|S`钸_̃Xg

	RxScalarField m_Grid;					//!< Obh

	// A֐l(XJ[l)擾pϐ(ǂ炩̂ݗp)
	const RXREAL* m_ptScalarField;				//!< XJ[lێTv{[
	RXREAL (*m_fpScalarFunc)(double, double, double, void*);	//!< XJ[lԂ֐|C^
	void *m_pScalarFuncPtr;

	RXREAL m_tIsoLevel;							//!< 臒l

	bool m_bValidSurface;					//!< bV̉


	// bV\zp̃e[u
	static const uint m_edgeTable[256];
	static const int m_triTable[256][16];



	// MARK:protectedo֐

	//! GbWID
	uint GetEdgeID(uint nX, uint nY, uint nZ, uint nEdgeNo);

	//! _ID
	uint GetVertexID(uint nX, uint nY, uint nZ);

	// GbW̓l_vZ
	RxVertexID CalculateIntersection(uint nX, uint nY, uint nZ, uint nEdgeNo);
	RxVertexID CalculateIntersectionF(uint nX, uint nY, uint nZ, uint nEdgeNo);

	//! ObhGbW[̉A֐l^Ԃœl_vZ
	RxVertexID Interpolate(double fX1, double fY1, double fZ1, double fX2, double fY2, double fZ2, RXREAL tVal1, RXREAL tVal2);

	//! _CbV􉽏o͌`Ŋi[
	void RenameVerticesAndTriangles(vector<Vec3> &vrts, uint &nvrts, vector<int> &tris, uint &ntris);

	//! _@vZ
	void CalculateNormals(const vector<Vec3> &vrts, uint nvrts, const vector<int> &tris, uint ntris, 
						  vector<Vec3> &nrms, uint &nnrms);

};


//-----------------------------------------------------------------------------
// rxMCMeshGPUNX
//-----------------------------------------------------------------------------
class rxMCMeshGPU
{
protected:
	// MC@p
	uint3 m_u3GridSize;				//!< Obh(nx,ny,nz)
	uint3 m_u3GridSizeMask;			//!< Obh/CfbNXϊ̃}XN
	uint3 m_u3GridSizeShift;		//!< Obh/CfbNXϊ̃Vtg

	float3 m_f3VoxelMin;			//!< Obhŏʒu
	float3 m_f3VoxelMax;			//!< Obhőʒu
	float3 m_f3VoxelH;				//!< Obh
	uint m_uNumVoxels;				//!< Obh
	uint m_uMaxVerts;				//!< ő咸_
	uint m_uNumActiveVoxels;		//!< bV݂{NZ
	uint m_uNumVrts;				//!< _
	uint m_uNumTris;				//!< bV

	// foCX
	float *g_dfVolume;				//!< A֐f[^i[Obh
	float *g_dfNoise;				//!< mCYli[Obh(`掞̐F肷̂Ɏgp)
	uint *m_duVoxelVerts;			//!< ObhɊ܂܂郁bV_
	uint *m_duVoxelVertsScan;		//!< ObhɊ܂܂郁bV_(Scan)
	uint *m_duCompactedVoxelArray;	//!< bV܂ރObh

	uint *m_duVoxelOccupied;		//!< |Sɑ݂{NZ̃Xg
	uint *m_duVoxelOccupiedScan;	//!< |Sɑ݂{NZ̃Xg(prefix scan)

#ifdef RX_CUMC_USE_GEOMETRY
	// 􉽏𐶐ƂɕKvȕϐ
	uint3 m_u3EdgeSize[3];			//!< GbW(nx,ny,nz)
	uint m_uNumEdges[4];			//!< GbW
#endif

#ifdef RX_CUMC_USE_GEOMETRY
	// 􉽏𐶐ƂɕKvȕϐ
	uint *m_duVoxelCubeIdx;			//!< Obh8_̉A֐l臒lȏォǂerbgɊi[ϐ

	uint *m_duEdgeOccupied;			//!< GbWɃbV_܂ނǂ(xCy, z̏)
	uint *m_duEdgeOccupiedScan;		//!< GbWɃbV_܂ނǂ(Scan)
	float *m_dfEdgeVrts;			//!< GbWƂɎZo_
	float *m_dfCompactedEdgeVrts;	//!< Ԃ߂_
	uint *m_duIndexArray;			//!< |S̊􉽏
	float *m_dfVertexNormals;		//!< _@
	uint *m_duVoxelTriNum;			//!< ObhƂ̎Op`bV
	uint *m_duVoxelTriNumScan;		//!< ObhƂ̎Op`bV(Scan)
#else
	// 􉽏KvƂȂƂ̂ݗp
	float4 *m_df4Vrts;				//!< |S_W
	float4 *m_df4Nrms;				//!< |S_@
#endif

	// zXg
	float4 *m_f4VertPos;			//!< _W
#ifdef RX_CUMC_USE_GEOMETRY
	uint3 *m_u3TriIdx;				//!< bVCfbNX
	uint *m_uScan;					//!< fobOp
#else
	float4 *m_f4VertNrm;			//!< _@
#endif

	int m_iVertexStore;
	bool m_bSet;


	// A֐l(XJ[l)擾pϐ(ǂ炩̂ݗp)
	const float* m_ptScalarField;				//!< XJ[lێTv{[
	float (*m_fpScalarFunc)(double, double, double);	//!< XJ[lԂ֐|C^

	float m_tIsoLevel;							//!< 臒l

	bool m_bValidSurface;					//!< bV̉

public:
	// RXgN^
	rxMCMeshGPU();

	// fXgN^
	~rxMCMeshGPU();
	
	//! A֐Op`bV𐶐
	bool CreateMesh(float (*func)(double, double, double), Vec3 min_p, double h, int n[3], float threshold, string method, 
					vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Op`bV𐶐
	bool CreateMeshV(Vec3 minp, double h, int n[3], float threshold, uint &nvrts, uint &ntris);
	
	//! z̊m
	bool Set(Vec3 vMin, Vec3 vH, int n[3], unsigned int vm);

	//! z̍폜
	void Clean(void);

	//! FBOɃf[^ݒ
	bool SetDataToFBO(GLuint uVrtVBO, GLuint uNrmVBO, GLuint uTriVBO);

	//! zXgzɃf[^ݒ
	bool SetDataToArray(vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Zbg
	void   SetSampleVolumeFromHost(float *hVolume);
	float* GetSampleVolumeDevice(void);

	void   SetSampleNoiseFromHost(float *hVolume);
	float* GetSampleNoiseDevice(void);

	// ő咸_
	uint GetMaxVrts(void){ return m_uMaxVerts; }

	// ő咸_vZpW
	void SetVertexStore(int vs){ m_iVertexStore = vs; }

	
#ifdef RX_CUMC_USE_GEOMETRY
	//! _f[^(foCX)
	float* GetVrtDev(void){ return (float*)m_dfCompactedEdgeVrts; }

	//! bVf[^(foCX)
	uint* GetIdxDev(void){ return (uint*)m_duIndexArray; }

	//! @f[^(foCX)
	float* GetNrmDev(void){ return (float*)m_dfVertexNormals; }

	//! _f[^(zXg)
	float GetVertex(int idx, int dim)
	{
		if(dim == 0){
			return m_f4VertPos[idx].x;
		}
		else if(dim == 1){
			return m_f4VertPos[idx].y;
		}
		else{
			return m_f4VertPos[idx].z;
		}
	}
	void GetVertex2(int idx, float *x, float *y, float *z)
	{
		*x = m_f4VertPos[idx].x;
		*y = m_f4VertPos[idx].y;
		*z = m_f4VertPos[idx].z;
	}

	//! bVf[^(zXg)
	void GetTriIdx(int idx, unsigned int *vidx0, unsigned int *vidx1, unsigned int *vidx2)
	{
		*vidx0 = m_u3TriIdx[idx].x;
		*vidx1 = m_u3TriIdx[idx].y;
		*vidx2 = m_u3TriIdx[idx].z;
	}
#else // #ifdef RX_CUMC_USE_GEOMETRY

	//! _f[^(foCX)
	float* GetVrtDev(void)
	{
		return (float*)m_df4Vrts;
	}

	//! @f[^(foCX)
	float* GetNrmDev(void)
	{
		return (float*)m_df4Nrms;
	}

#endif // #ifdef RX_CUMC_USE_GEOMETRY
};





/*!
 * bV(Tv{[쐬)
 * @param[in] sf Obh
 * @param[in] func A֐l擾֐|C^
 * @param[in] threshold 臒l
 * @param[out] vrts bV_
 * @param[out] nrms bV_@
 * @param[out] tris bV􉽏(_ڑ)
 */
static void GenerateValueArray(RXREAL **field, RXREAL (*func)(void*, double, double, double), void* func_ptr, const RxScalarField sf)
{
	int nx, ny, nz;
	nx = sf.iNum[0]+1;
	ny = sf.iNum[1]+1;
	nz = sf.iNum[2]+1;

	Vec3 minp = sf.fMin;
	Vec3 d = sf.fWidth;

	if(*field) delete [] *field;
	*field = new RXREAL[nx*ny*nz];
	for(int k = 0; k < nz; ++k){
		for(int j = 0; j < ny; ++j){
			for(int i = 0; i < nx; ++i){
				int idx = k*nx*ny+j*nx+i;
				Vec3 pos = minp+Vec3(i, j, k)*d;

				RXREAL val = func(func_ptr, pos[0], pos[1], pos[2]);
				(*field)[idx] = val;
			}
		}
	}
}



#endif // _RX_MC_MESH_H_

