/*!
  @file rx_particle_on_surf.h
	
  @brief A֐\ʂɃp[eBNzu
 
  @author Makoto Fujisawa
  @date   2013-06
*/


#ifndef _RX_PARTICLE_ON_SURFACE_H_
#define _RX_PARTICLE_ON_SURFACE_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cmath>

// STL
#include <vector>
#include <string>

#include <iostream>

#include "rx_utility.h"
#include "rx_nnsearch.h"


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
using namespace std;
typedef unsigned int uint;

#ifndef RXREAL
	#define RXREAL float
#endif
#ifndef DIM
	#define DIM 4
#endif





//-----------------------------------------------------------------------------
// A֐lʂɃp[eBNzuNX
//  -Ql: A. Witkin and P. Heckbert, "Using particles to sample and control implicit surfaces", SIGGRAPH1994.
//-----------------------------------------------------------------------------
class rxParticleOnSurf
{
	struct rxSurfParticle
	{
		Vec3 Vel;
		Vec3 Ep;
		RXREAL Sigma;
		int Flag;

		rxSurfParticle() : Vel(Vec3(0.0)), Ep(Vec3(0.0)), Sigma(0.0), Flag(0) {}
	};

protected:
	uint m_uNumParticles;

	RXREAL m_fParticleRadius;
	RXREAL m_fEffectiveRadius;		//!< La
	RXREAL m_fKernelRadius;			//!< J[l̉e͈

	vector<RXREAL> m_vPos;			//!< ߖTT[`ɓn߂ɈʒuʊǗ
	vector<rxSurfParticle> m_vPs;	//!< p[eBN

	// Repulsionp[^
	RXREAL m_fAlpha;	//!< repulsion amplitude
	RXREAL m_fEh;		//!< desired energy
	RXREAL m_fRho;		//!< p[eBNenergym_fEhɕۂ߂̌W
	RXREAL m_fPhi;		//!< p[eBNʒuȖʏɕۂ߂̌W
	RXREAL m_fBeta;		//!< zero-divideh~p
	RXREAL m_fSh;		//!< desired repulsion radius ([Uw)
	RXREAL m_fSmax;		//!< maximum repulsion radius

	RXREAL m_fGamma;	//!< equilibrium speed (Ђɑ΂{)
	RXREAL m_fNu;		//!< p[eBN̂߂̌W
	RXREAL m_fDelta;	//!< p[eBN폜̂߂̌W

	void *m_pFuncPtr;
	Vec4 (*m_fpFunc)(void*, double, double, double);

	rxNNGrid *m_pNNGrid;			//!< ObhɂߖTT
	vector< vector<rxNeigh> > m_vNeighs;	//!< ߖTp[eBN

public:
	//! RXgN^
	rxParticleOnSurf();

	//! fXgN^
	~rxParticleOnSurf(){}

	// p[eBN&j
	void Initialize(const vector<Vec3> &vrts, double rad, Vec3 minp, Vec3 maxp, 
					Vec4 (*func)(void*, double, double, double), void* func_ptr);
	void Finalize(void);

	//! p[eBN
	int	GetNumParticles() const { return m_uNumParticles; }

	//! p[eBNa
	float GetParticleRadius(void){ return m_fParticleRadius; }

	//! p[eBNf[^̎擾
	RXREAL* GetPositionArray(void){ return &m_vPos[0]; }

	//! p[eBNʒu̍XV
	void Update(double dt, int &num_iter, RXREAL &eps);

protected:
	//! Ђ̍XV
	void updateSigma(double dt);

	//! ͂̌vZ
	void applyRepulsion(double dt);
	void applyRepulsion2(double dt);

	//! \ʂɉړ
	void applyFloating(double dt, double &v_avg);

	//!< p[eBN/j̔
	void testFissionDeath(void);

public:
	// ߖT擾
	void GetNearestNeighbors(Vec3 pos, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighbors(int idx, RXREAL *p, vector<rxNeigh> &neighs, RXREAL h = -1.0);

	// ZɃp[eBNi[
	void SetParticlesToCell(void);
	void SetParticlesToCell(RXREAL *prts, int n, RXREAL h);

};


#endif // _RX_PARTICLE_ON_SURFACE_H_

