/*! 
  @file rx_sph.cu
	
  @brief CUDAɂSPH

  @author Makoto Fujisawa
  @date 2009-08, 2011-06
*/
// FILE --rx_sph.cu--


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cstdio>
#include <GL/glew.h>
#include <GL/glut.h>

#include "rx_pbf_kernel.cu"

#include <thrust/device_vector.h>
#include <thrust/scan.h>



//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
extern "C"
{
	
void CuSetParameters(rxSimParams *hostParams)
{
	// copy parameters to constant memory
	RX_CUCHECK( cudaMemcpyToSymbol(params, hostParams, sizeof(rxSimParams)) );
}

void CuClearData(void)
{
}

/*!
 * thrust::exclusive_scaňĂяo
 * @param[out] dScanData scañf[^
 * @param[in] dData f[^
 * @param[in] num f[^
 */
void CuScanf(float* dScanData, float* dData, unsigned int num)
{
	thrust::exclusive_scan(thrust::device_ptr<float>(dData), 
						   thrust::device_ptr<float>(dData+num),
						   thrust::device_ptr<float>(dScanData));
}


/*!
 * Z̃nbVvZ
 * @param[in] 
 * @return 
 */
void CuCalcHash(uint* dGridParticleHash, uint* dSortedIndex, float* dPos, int nprts)
{
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	calcHashD<<< numBlocks, numThreads >>>(dGridParticleHash,
										   dSortedIndex,
										   (float4*)dPos,
										   nprts);
	
	RX_CUERROR("calcHashD kernel execution failed");	// J[lG[`FbN
}


/*!
 * p[eBNz\[gꂽԂɕёւC
 * eZ̎n܂ƏĨCfbNX
 * @param[in] cell p[eBNObhf[^
 * @param[in] oldPos p[eBNʒu
 * @param[in] oldVel p[eBNx
 */
void CuReorderDataAndFindCellStart(rxParticleCell cell, float* oldPos, float* oldVel)
{
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	RX_CUCHECK(cudaMemset(cell.dCellStart, 0xffffffff, cell.uNumCells*sizeof(uint)));

	uint smemSize = sizeof(uint)*(numThreads+1);

	// J[ls
	reorderDataAndFindCellStartD<<< numBlocks, numThreads, smemSize>>>(cell, (float4*)oldPos, (float4*)oldVel);

	RX_CUERROR("reorderDataAndFindCellStartD kernel execution failed");
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * Z̃nbVvZ
 * @param[in] 
 * @return 
 */
void CuCalcHashB(uint* dGridParticleHash, uint* dSortedIndex, float* dPos, 
				 float3 world_origin, float3 cell_width, uint3 grid_size, int nprts)
{
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	calcHashB<<< numBlocks, numThreads >>>(dGridParticleHash,
										   dSortedIndex,
										   (float4*)dPos,
										   world_origin, 
										   cell_width, 
										   grid_size, 
										   nprts);
	
	RX_CUERROR("Kernel execution failed : calcHashB");	// J[lG[`FbN
}

/*!
 * p[eBNz\[gꂽԂɕёւC
 * eZ̎n܂ƏĨCfbNX
 * @param[in] cell p[eBNObhf[^
 * @param[in] oldPos p[eBNʒu
 */
void CuReorderDataAndFindCellStartB(rxParticleCell cell, float* oldPos)
{
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	RX_CUCHECK(cudaMemset(cell.dCellStart, 0xffffffff, cell.uNumCells*sizeof(uint)));

	uint smemSize = sizeof(uint)*(numThreads+1);

	// J[ls
	reorderDataAndFindCellStartB<<< numBlocks, numThreads, smemSize>>>(cell, (float4*)oldPos);

	RX_CUERROR("Kernel execution failed: CuReorderDataAndFindCellStartB");
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}










//-----------------------------------------------------------------------------
// Ep[eBN
//-----------------------------------------------------------------------------
/*!
 * Ep[eBN̑̐ςvZ
 *  - "Versatile Rigid-Fluid Coupling for Incompressible SPH", 2.2 (3)̏
 * @param[out] dVolB Ep[eBN̑̐
 * @param[in]  mass p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
void CuSphBoundaryVolume(float* dVolB, float mass, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalBoundaryVolume<<< numBlocks, numThreads >>>(dVolB, cell);

	RX_CUERROR("kernel execution failed : sphCalBoundaryVolume");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNx̌vZ(J[lĂяo)
 * @param[out] dDens p[eBNx
 * @param[out] dPres p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
void CuSphBoundaryDensity(float* dDens, float* dPres, float* dPos, float* dVolB, rxParticleCell bcell, uint pnum)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(pnum, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalBoundaryDensity<<< numBlocks, numThreads >>>(dDens, dPres, (float4*)dPos, dVolB, bcell, pnum);

	RX_CUERROR("kernel execution failed : sphCalBoundaryDensity");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNɂ͂̌vZ(J[lĂяo)
 * @param[in] dDens p[eBNx
 * @param[in] dPres p[eBN
 * @param[out] dFrc p[eBNɂ
 * @param[in] cell p[eBNObhf[^
 * @param[in] dt ԃXebv
 */
void CuSphBoundaryForces(float* dDens, float* dPres, float* dPos, float* dVolB, float* dFrc, rxParticleCell bcell, uint pnum)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(pnum, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	sphCalBoundaryForce<<< numBlocks, numThreads >>>(dDens, dPres, (float4*)dPos, dVolB, (float4*)dFrc, bcell, pnum);

	RX_CUERROR("kernel execution failed : sphCalBoundaryForce");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}





//-----------------------------------------------------------------------------
// PBF
//-----------------------------------------------------------------------------

/*!
 * p[eBNx̌vZ(J[lĂяo)
 * @param[out] dDens p[eBNx
 * @param[out] dPres p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
void CuPbfDensity(float* dDens, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalDensity<<< numBlocks, numThreads >>>(dDens, cell);

	RX_CUERROR("pbfCalDensity kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNɂ͂̌vZ(J[lĂяo)
 * @param[in] dDens p[eBNx
 * @param[out] dFrc p[eBNɂ
 * @param[in] cell p[eBNObhf[^
 * @param[in] dt ԃXebv
 */
void CuPbfExternalForces(float* dDens, float* dFrc, rxParticleCell cell, float dt)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalExternalForces<<< numBlocks, numThreads >>>(dDens, (float4*)dFrc, cell);

	RX_CUERROR("pbfCalExternalForces kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * XP[Ot@N^̌vZ
 * @param[in] dPos p[eBNSW
 * @param[out] dDens p[eBNx
 * @param[out] dScl XP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in] cell p[eBNObhf[^
 */
void CuPbfScalingFactor(float* dPos, float* dDens, float* dScl, float eps, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalScalingFactor<<< numBlocks, numThreads >>>((float4*)dPos, dDens, dScl, eps, cell);

	RX_CUERROR("pbfCalScalingFactor kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * ϖxϓ̌vZ
 *  - ׂẴp[eBNx̏xƂ̍J[lŌvZCPrefix Sum (Scan)ł̍v߂
 * @param[out] dErrScan ϓlScanʂi[z
 * @param[out] dErr p[eBNxϓl
 * @param[in] dDens p[eBNx
 * @param[in] rest_dens x
 * @param[in] nprts p[eBN
 * @return ϖxϓ
 */
float CuPbfCalDensityFluctuation(float* dErrScan, float* dErr, float* dDens, float rest_dens, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfDensityFluctuation<<< numBlocks, numThreads >>>(dErr, dDens, rest_dens, nprts);

	RX_CUERROR("pbfDensityFluctuation kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂

	// ep[eBN̖xϓScan
	CuScanf(dErrScan, dErr, nprts);

	// Exclusive scan (Ō̗vf0Ԗڂn-2Ԗڂ܂ł̍vɂȂĂ)Ȃ̂ŁC
	// ScanOz̍Ō(n-1Ԗ)̗vfƍv邱ƂŖxϓ̍vvZ
	float lval, lsval;
	RX_CUCHECK(cudaMemcpy((void*)&lval, (void*)(dErr+nprts-1), sizeof(float), cudaMemcpyDeviceToHost));
	RX_CUCHECK(cudaMemcpy((void*)&lsval, (void*)(dErrScan+nprts-1), sizeof(float), cudaMemcpyDeviceToHost));
	float dens_var = lval+lsval;

	return dens_var/(float)nprts;
}

/*!
 * ʒuCʂ̌vZ
 * @param[in] dPos p[eBNSW
 * @param[in] dScl XP[Ot@N^
 * @param[out] dDp ʒuC
 * @param[in] cell p[eBNObhf[^
 */
void CuPbfPositionCorrection(float* dPos, float* dScl, float* dDp, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfPositionCorrection<<< numBlocks, numThreads >>>((float4*)dPos, dScl, (float4*)dDp, cell);

	RX_CUERROR("pbfPositionCorrection kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNʒuXV
 * @param[inout] dPos p[eBNʒu
 * @param[in] dDp ʒuC
 * @param[in] nprts p[eBN
 */
void CuPbfCorrectPosition(float* dPos, float* dDp, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCorrectPosition<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dDp, nprts);
	
	RX_CUERROR("pbfCorrectPosition kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}



/*!
 * Ep[eBNx]̃p[eBNxɉ
 * @param[inout] dDens ̃p[eBNx
 * @param[in] dPos  ̃p[eBN
 * @param[in] dVolB Ep[eBN̐
 * @param[in] bcell Ep[eBNObhf[^
 */
void CuPbfBoundaryDensity(float* dDens, float* dPos, float* dVolB, rxParticleCell bcell, uint pnum)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(pnum, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalBoundaryDensity<<< numBlocks, numThreads >>>(dDens, (float4*)dPos, dVolB, bcell, pnum);

	RX_CUERROR("pbfCalBoundaryDensity kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * XP[Ot@N^̌vZ(Ep[eBN܂)
 * @param[in] dPos ̃p[eBNSW
 * @param[out] dDens ̃p[eBNx
 * @param[out] dScl ̃p[eBÑXP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in] cell ̃p[eBNObhf[^
 * @param[in] dVolB Ep[eBN̐
 * @param[out] dSclB Ep[eBÑXP[Ot@N^
 * @param[in] bcell Ep[eBNObhf[^
 */
void CuPbfScalingFactorWithBoundary(float* dPos, float* dDens, float* dScl, float eps, rxParticleCell cell, 
									   float* dVolB, float* dSclB, rxParticleCell bcell)
{
	// ̃p[eBN̐Xbh𗧂Ă
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalScalingFactorWithBoundary<<< numBlocks, numThreads >>>((float4*)dPos, dDens, dScl, eps, cell, dVolB, bcell);

	RX_CUERROR("pbfCalScalingFactorWithBoundary kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂

	// Ep[eBÑXP[Ot@N^̌vZ
	// Ep[eBN̐Xbh𗧂Ă
	computeGridSize(bcell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalBoundaryScalingFactor<<< numBlocks, numThreads >>>((float4*)dPos, dDens, eps, cell, dVolB, dSclB, bcell);

	RX_CUERROR("pbfCalScalingFactorWithBoundary kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂

}

/*!
 * ʒuCʂ̌vZ(Ep[eBN܂)
 * @param[in] dPos ̃p[eBNSW
 * @param[in] dScl ̃p[eBÑXP[Ot@N^
 * @param[out] dDens ̃p[eBNʒuC
 * @param[in] cell ̃p[eBNObhf[^
 * @param[in] dVolB Ep[eBN̐
 * @param[in] dSclB Ep[eBÑXP[Ot@N^
 * @param[in] bcell Ep[eBNObhf[^
 */
void CuPbfPositionCorrectionWithBoundary(float* dPos, float* dScl, float* dDp, rxParticleCell cell, 
											float* dVolB, float* dSclB, rxParticleCell bcell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfPositionCorrectionWithBoundary<<< numBlocks, numThreads >>>((float4*)dPos, dScl, (float4*)dDp, cell, 
																	  dVolB, dSclB, bcell);

	RX_CUERROR("pbfPositionCorrectionWithBoundary kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}



/*!
 * p[eBNʒuCx̍XV
 *  - Op`|SE
 *  - ݂̈ʒûݎgďՓ˔
 * @param[in] dPos p[eBNʒu
 * @param[in] dVel p[eBNx
 * @param[in] dAcc p[eBNx
 * @param[out] dNewPos XVꂽp[eBNʒu
 * @param[out] dNewVel XVꂽp[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
void CuPbfIntegrate(float* dPos, float* dVel, float* dAcc, 
					float* dNewPos, float* dNewVel, float dt, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfIntegrate<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dVel, (float4*)dAcc, 
												 (float4*)dNewPos, (float4*)dNewVel, dt, nprts);
	
	RX_CUERROR("pbfIntegrate kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNʒuCx̍XV
 *  - Op`|SE
 *  - ݂̈ʒûݎgďՓ˔
 * @param[in] dPos p[eBNʒu
 * @param[in] dVel p[eBNx
 * @param[in] dAcc p[eBNx
 * @param[out] dNewPos XVꂽp[eBNʒu
 * @param[out] dNewVel XVꂽp[eBNx
 * @param[in] dVrts Op`|S_
 * @param[in] dTris Op`|SCfbNX
 * @param[in] tri_num Op`|S
 * @param[in] dt ԃXebv
 * @param[in] cell p[eBNObhf[^
 */
void CuPbfIntegrateWithPolygon(float* dPos, float* dVel, float* dAcc, 
							   float* dNewPos, float* dNewVel, 
							   float* dVrts, int* dTris, int tri_num, float dt, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfIntegrateWithPolygon<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dVel, (float4*)dAcc, 
															(float4*)dNewPos, (float4*)dNewVel, (float3*)dVrts, (int3*)dTris, tri_num, dt, cell);
	
	RX_CUERROR("pbfIntegrateWithPolygon kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * p[eBNʒuCx̍XV
 *  - ݂̈ʒuƏC̈ʒu2gďՓ˔(PBFɎgp)
 * @param[in] dPos p[eBNʒu
 * @param[in] dVel p[eBNx
 * @param[in] dAcc p[eBNx
 * @param[out] dNewPos XVꂽp[eBNʒu
 * @param[out] dNewVel XVꂽp[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
void CuPbfIntegrate2(float* dPos, float* dVel, float* dAcc, 
					 float* dNewPos, float* dNewVel, float dt, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfIntegrate2<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dVel, (float4*)dAcc, 
												  (float4*)dNewPos, (float4*)dNewVel, dt, nprts);
	
	RX_CUERROR("pbfIntegrate2 kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}
/*!
 * p[eBNʒuCx̍XV
 *  - Op`|SE
 *  - ݂̈ʒuƏC̈ʒu2gďՓ˔(PBFɎgp)
 * @param[in] dPos p[eBNʒu
 * @param[in] dVel p[eBNx
 * @param[in] dAcc p[eBNx
 * @param[out] dNewPos XVꂽp[eBNʒu
 * @param[out] dNewVel XVꂽp[eBNx
 * @param[in] dVrts Op`|S_
 * @param[in] dTris Op`|SCfbNX
 * @param[in] tri_num Op`|S
 * @param[in] dt ԃXebv
 * @param[in] cell p[eBNObhf[^
 */
void CuPbfIntegrateWithPolygon2(float* dPos, float* dVel, float* dAcc, 
								float* dNewPos, float* dNewVel, 
								float* dVrts, int* dTris, int tri_num, float dt, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfIntegrateWithPolygon2<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dVel, (float4*)dAcc, 
															 (float4*)dNewPos, (float4*)dNewVel, (float3*)dVrts, (int3*)dTris, tri_num, dt, cell);
	
	RX_CUERROR("pbfIntegrateWithPolygon2 kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}


/*!
 * p[eBNʒuCx̍XV
 * @param[in] pos XVꂽp[eBNʒu
 * @param[inout] new_pos Xebvŏ̃p[eBNʒu/Vp[eBNx
 * @param[out] new_vel Vp[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
void CuPbfUpdatePosition(float* dPos, float* dNewPos, float* dNewVel, float dt, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfUpdatePosition<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dNewPos, (float4*)dNewVel, dt, nprts);
	
	RX_CUERROR("CupbfUpdatePosition kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBNʒuCx̍XV
 * @param[in] pos XVꂽp[eBNʒu
 * @param[inout] new_pos Xebvŏ̃p[eBNʒu/Vp[eBNx
 * @param[out] new_vel Vp[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
void CuPbfUpdateVelocity(float* dPos, float* dNewPos, float* dNewVel, float dt, uint nprts)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(nprts, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfUpdateVelocity<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dNewPos, (float4*)dNewVel, dt, nprts);
	
	RX_CUERROR("pbfUpdateVelocity kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * XSPHɂSvZ
 * @param[in] dPos p[eBNSW
 * @param[in] dVel p[eBNx
 * @param[out] dNewVel XVꂽp[eBNx
 * @param[in] dDens p[eBNx
 * @param[in] c SvZpp[^
 * @param[in] cell p[eBNObhf[^
 */
void CuXSphViscosity(float* dPos, float* dVel, float* dNewVel, float* dDens, float c, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	xsphVisocosity<<< numBlocks, numThreads >>>((float4*)dPos, (float4*)dVel, (float4*)dNewVel, dDens, c, cell);

	RX_CUERROR("xsphVisocosity kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * Obh̖xZo
 * @param[out] dGridD Obh̖xl
 * @param[in] cell p[eBNObhf[^
 * @param[in] nx,ny,nz Obh
 * @param[in] x0,y0,z0 ObhŏW
 * @param[in] dx,dy,dz Obh
 */
void CuPbfGridDensity(float *dGridD, rxParticleCell cell, 
					  int nx, int ny, int nz, float x0, float y0, float z0, float dx, float dy, float dz)
{
	uint3  gnum = make_uint3(nx, ny, nz);
	float3 gmin = make_float3(x0, y0, z0);
	float3 glen = make_float3(dx, dy, dz);

	int numcell = gnum.x*gnum.y*gnum.z;

	int threads = 128;
	dim3 grid((numcell+threads-1)/threads, 1, 1);
	if(grid.x > 65535){
		grid.y = (grid.x+32768-1)/32768;
		grid.x = 32768;
	}

	// J[ls
	pbfCalDensityInGrid<<<grid, threads>>>(dGridD, cell, gnum, gmin, glen);

	RX_CUERROR("pbfCalDensityInGrid Kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂
}

/*!
 * p[eBN@̌vZ
 * @param[out] dNrms p[eBN@
 * @param[int] dDens p[eBNx
 * @param[in]  cell p[eBNObhf[^
 */
void CuPbfNormal(float* dNrms, float* dDens, rxParticleCell cell)
{
	// 1Xbh/p[eBN
	uint numThreads, numBlocks;
	computeGridSize(cell.uNumParticles, THREAD_NUM, numBlocks, numThreads);

	// J[ls
	pbfCalNormal<<< numBlocks, numThreads >>>((float4*)dNrms, dDens, cell);

	RX_CUERROR("pbfCalNormal kernel execution failed");	// J[lsG[`FbN
	RX_CUCHECK(cudaThreadSynchronize());		// SẴXbhÎ҂

}










}   // extern "C"
