/*! 
  @file rx_pvsph_kernel.cu
	
  @brief CUDAɂPVSPH
 
  @author Makoto Fujisawa
  @date 2014-12
*/

#ifndef _RX_PVSPH_KERNEL_CU_
#define _RX_PVSPH_KERNEL_CU_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_cu_common.cu"

//-----------------------------------------------------------------------------
// nbV
//-----------------------------------------------------------------------------
/*!
 * ep[eBÑObhnbVľvZ
 * @param[out] gridParticleHash nbVl
 * @param[out] dSortedIndex p[eBNCfbNXl(\[gȌl)
 * @param[in] pos p[eBNʒui[z
 * @param[in] nprts p[eBN
 */
__global__
void calcHashD(uint*   dGridParticleHash, 
			   uint*   dSortedIndex, 
			   float4* dPos, 
			   uint	   nprts)
{
	uint index = __umul24(blockIdx.x, blockDim.x)+threadIdx.x;
	if(index >= nprts) return;
	
	volatile float4 p = dPos[index];
	int3 gridPos = calcGridPos(make_float3(p.x, p.y, p.z));
	uint hash = calcGridHash(gridPos);

	dGridParticleHash[index] = hash;
	dSortedIndex[index] = index;
}

/*!
 * ep[eBÑObhnbVl
 *  - paramsgȂŃObhŗ^
 * @param[out] gridParticleHash p[eBÑObhnbVl
 * @param[out] dSortedIndex p[eBNCfbNXl(\[gȌl)
 * @param[in] dPos p[eBNʒui[z
 * @param[in] world_origin ObhŏWl
 * @param[in] cell_width Obh̃Z
 * @param[in] grid_size Obh
 * @param[in] nprts p[eBN
 */
__global__
void calcHashB(uint*   dGridParticleHash, 
			   uint*   dSortedIndex, 
			   float4*  dPos, 
			   float3  world_origin, 
			   float3  cell_width, 
			   uint3   grid_size, 
			   uint	   nprts)
{
	uint index = __umul24(blockIdx.x, blockDim.x)+threadIdx.x;
	if(index >= nprts) return;
	
	float3 p = make_float3(dPos[index]);
	int3 gridPos = calcGridPosB(make_float3(p.x, p.y, p.z), world_origin, cell_width, grid_size);
	uint hash = calcGridHashB(gridPos, grid_size);

	dGridParticleHash[index] = hash;
	dSortedIndex[index] = index;
}


/*!
 * p[eBNf[^\[gāCnbV̊eZ̍ŏ̃AhX
 * @param[in] cell p[eBNObhf[^
 * @param[in] dSortedPos p[eBNʒu
 * @param[in] dSortedVel p[eBNx
 */
__global__
void reorderDataAndFindCellStartD(rxParticleCell cell, float4* dSortedPos, float4* dSortedVel)
{
	extern __shared__ uint sharedHash[];	// TCY : blockSize+1
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	
	uint hash;
	if(index < cell.uNumParticles){
		hash = cell.dGridParticleHash[index];	// nbVl

		sharedHash[threadIdx.x+1] = hash;	// nbVlVFA[hɊi[

		if(index > 0 && threadIdx.x == 0){
			// eVFA[h̍ŏׂ͗̃Obh̃p[eBÑnbVli[
			sharedHash[0] = cell.dGridParticleHash[index-1];
		}
	}

	__syncthreads();
	
	if(index < cell.uNumParticles){
		// CfbNX0łĆCÕp[eBÑObhnbVlقȂꍇC
		// p[eBN͕̈̍ŏ
		if(index == 0 || hash != sharedHash[threadIdx.x]){
			cell.dCellStart[hash] = index;
			if(index > 0){
				// Õp[eBŃCÖ̍̕Ō
				cell.dCellEnd[sharedHash[threadIdx.x]] = index;
			}
		}

		// CfbNXŌȂ΁C̈̍Ō
		if(index == cell.uNumParticles-1){
			cell.dCellEnd[hash] = index+1;
		}

		// ʒuƑx̃f[^ёւ
		// \[gCfbNXŎQƂ\T̃O[oANZXɗ͗}邽߂Ƀf[^̂̂ёւ
		uint sortedIndex = cell.dSortedIndex[index];
		float4 pos = FETCH(dSortedPos, sortedIndex);
		float4 vel = FETCH(dSortedVel, sortedIndex);

		cell.dSortedPos[index] = pos;
		cell.dSortedVel[index] = vel;
	}
}

/*!
 * p[eBNf[^\[gāCnbV̊eZ̍ŏ̃AhX
 *  - ʒû
 * @param[in] cell p[eBNObhf[^
 * @param[in] dPos p[eBNʒu
 */
__global__
void reorderDataAndFindCellStartB(rxParticleCell cell, float4* dPos)
{
	extern __shared__ uint sharedHash[];	// TCY : blockSize+1
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	
	uint hash;
	if(index < cell.uNumParticles){
		hash = cell.dGridParticleHash[index];	// nbVl

		sharedHash[threadIdx.x+1] = hash;	// nbVlVFA[hɊi[

		if(index > 0 && threadIdx.x == 0){
			// eVFA[h̍ŏׂ͗̃Obh̃p[eBÑnbVli[
			sharedHash[0] = cell.dGridParticleHash[index-1];
		}
	}

	__syncthreads();
	
	if(index < cell.uNumParticles){
		// CfbNX0łĆCÕp[eBÑObhnbVlقȂꍇC
		// p[eBN͕̈̍ŏ
		if(index == 0 || hash != sharedHash[threadIdx.x]){
			cell.dCellStart[hash] = index;
			if(index > 0){
				// Õp[eBŃCÖ̍̕Ō
				cell.dCellEnd[sharedHash[threadIdx.x]] = index;
			}
		}

		// CfbNXŌȂ΁C̈̍Ō
		if(index == cell.uNumParticles-1){
			cell.dCellEnd[hash] = index+1;
		}

		// ʒuƑx̃f[^ёւ
		// \[gCfbNXŎQƂ\T̃O[oANZXɗ͗}邽߂Ƀf[^̂̂ёւ
		uint sortedIndex = cell.dSortedIndex[index];
		float4 pos = dPos[sortedIndex];
		cell.dSortedPos[index] = pos;
	}
}



//-----------------------------------------------------------------------------
// Ep[eBNJ[l
//-----------------------------------------------------------------------------
/*!
 * ^ꂽZ̃p[eBNƂ̋狫Ep[eBN̑̐ςvZ
 * @param[in] gridPos Obhʒu
 * @param[in] index p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calBoundaryVolumeCell(int3 gridPos, uint i, float3 pos0, rxParticleCell cell)
{
	uint gridHash = calcGridHashB(gridPos, params.GridSizeB);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = cell.dCellStart[gridHash];

	float h = params.EffectiveRadius;
	float mw = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = cell.dCellEnd[gridHash];
		for(uint j = startIndex; j < endIndex; ++j){
			float3 pos1 = make_float3(cell.dSortedPos[j]);

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h){
				float q = h*h-r*r;
				mw += params.Mass*params.Wpoly6*q*q*q;
			}
		}
	}

	return mw;
}

/*!
 * Ep[eBN̑̐όvZ(J[l֐)
 * @param[out] newVolB p[eBN̐
 * @param[in]  cell Ep[eBNObhf[^
 */
__global__
void sphCalBoundaryVolume(float* newVolB, rxParticleCell cell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(cell.dSortedPos[index]);	// p[eBNʒu
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu
	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos1 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);


	// ͂̃Obh܂߂ċߖTT
	float mw = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				mw += calBoundaryVolumeCell(n_grid_pos, index, pos, cell);
			}
		}
	}

	// ̐ςʂɏ
	uint oIdx = cell.dSortedIndex[index];
	newVolB[oIdx] = params.Mass/mw;
}

/*!
 * ^ꂽZ̃p[eBNƂ̋疧xvZ
 * @param[in] gridPos Obhʒu
 * @param[in] i p[eBNCfbNX
 * @param[in] pos0 vZW
 * @param[in] dVolB Ep[eBNz̐
 * @param[in] bcell Ep[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calBoundaryDensityCell(int3 gridPos, uint i, float3 pos0, float* dVolB, rxParticleCell bcell)
{
	uint gridHash = calcGridHashB(gridPos, params.GridSizeB);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = bcell.dCellStart[gridHash];

	float h = params.EffectiveRadius;
	float dens = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = bcell.dCellEnd[gridHash];
		for(uint j = startIndex; j < endIndex; ++j){
			//if(j == i) continue;

			float3 pos1 = make_float3(bcell.dSortedPos[j]);
			uint jdx = bcell.dSortedIndex[j];

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h){
				float q = h*h-r*r;
				dens += params.Density*dVolB[jdx]*params.Wpoly6*q*q*q;
			}
		}
	}

	return dens;
}

/*!
 * Ep[eBNxvZ(J[l֐)
 * @param[out] newDens Ep[eBNx
 * @param[out] newPres Ep[eBN - PBFł͎gȂ
 * @param[in] dPos  Ep[eBNʒu
 * @param[in] dVolB Ep[eBNz̐
 * @param[in] bcell Ep[eBNObhf[^
 * @param[in] pnum  Ep[eBN
 */
__global__
void sphCalBoundaryDensity(float* newDens, float* newPres, float4* dPos, float* dVolB, rxParticleCell bcell, uint pnum)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= pnum) return;	
	
	float3 pos = make_float3(dPos[index]);	// p[eBNʒu
	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos1 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);

	// ͂̃Obh܂߂ċߖTTCxvZ
	float dens = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calBoundaryDensityCell(n_grid_pos, index, pos, dVolB, bcell);
			}
		}
	}

	dens += newDens[index];

	// KX萔g͎Zo
	float pres;
	pres = params.GasStiffness*(dens-params.Density);

	// xƈ͒lʂɏ
	newDens[index] = dens;
	newPres[index] = pres;
}


/*!
 * ^ꂽZ̃p[eBNƂ̋疧xvZ
 * @param[in] gridPos Obhʒu
 * @param[in] i p[eBNCfbNX
 * @param[in] pos0 vZW
 * @param[in] dVolB Ep[eBNz̐
 * @param[in] dens0 p[eBNi̖x
 * @param[in] pres0 p[eBNï
 * @param[in] bcell Ep[eBNObhf[^
 * @return Z̃p[eBNvZO
 */
__device__
float3 calBoundaryForceCell(int3 gridPos, uint i, float3 pos0, float* dVolB, float dens0, float pres0, rxParticleCell bcell)
{
	uint gridHash = calcGridHashB(gridPos, params.GridSizeB);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = bcell.dCellStart[gridHash];

	float h = params.EffectiveRadius;
	float3 bp = make_float3(0.0f);
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = bcell.dCellEnd[gridHash];
		float prsi = pres0/(dens0*dens0);
		for(uint j = startIndex; j < endIndex; ++j){
			float3 pos1 = make_float3(bcell.dSortedPos[j]);
			uint jdx = bcell.dSortedIndex[j];

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h && r > 0.0001){
				float q = h-r;
				bp += -params.Density*dVolB[jdx]*prsi*params.GWspiky*q*q*rij/r;
			}
		}
	}

	return bp;
}

/*!
 * Ep[eBNɂ͂̌vZ(J[l֐)
 * @param[in] dDens Ep[eBNx
 * @param[in] dPres Ep[eBN
 * @param[in] dPos  Ep[eBNʒu
 * @param[in] dVolB Ep[eBNz̐
 * @param[out] outFrc O
 * @param[in] bcell Ep[eBNObhf[^
 * @param[in] pnum  Ep[eBN
 */
__global__
void sphCalBoundaryForce(float* dDens, float* dPres, float4* dPos, float* dVolB, float4* outFrc, rxParticleCell bcell, uint pnum)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= pnum) return;	
	
	float3 pos = make_float3(dPos[index]);	// p[eBNʒu
	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos1 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);

	// xƈ
	float dens0 = dDens[index];
	float pres0 = dPres[index];

	// ͂̃Obh܂߂ċߖTTCxvZ
	float3 frc = make_float3(0.0f);
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				frc += calBoundaryForceCell(n_grid_pos, index, pos, dVolB, dens0, pres0, bcell);
			}
		}
	}

	// xƈ͒lʂɏ
	outFrc[index] += make_float4(frc, 0.0f);
}


//-----------------------------------------------------------------------------
// pbf
//-----------------------------------------------------------------------------
/*!
 * ^ꂽZ̃p[eBNƂ̋疧xvZ
 * @param[in] gridPos Obhʒu
 * @param[in] i p[eBNCfbNX
 * @param[in] pos0 vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calDensityCellPB(int3 gridPos, uint i, float3 pos0, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float h = params.EffectiveRadius;
	float dens = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);
		for(uint j = startIndex; j < endIndex; ++j){
			//if(j == i) continue;

			float3 pos1 = make_float3(FETCHC(dSortedPos, j));

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h){
				float q = h*h-r*r;
				dens += params.Mass*params.Wpoly6*q*q*q;
			}
		}
	}

	return dens;
}



/*!
 * p[eBNxvZ(J[l֐)
 * @param[out] newDens p[eBNx
 * @param[in]  cell p[eBNObhf[^
 */
__global__
void pbfCalDensity(float* newDens, rxParticleCell cell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu
	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ͂̃Obh܂߂ċߖTTCxvZ
	float dens = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calDensityCellPB(n_grid_pos, index, pos, cell);
			}
		}
	}

	// xƈ͒lʂɏ
	uint oIdx = cell.dSortedIndex[index];
	newDens[oIdx] = dens;
}

/*!
 * ^ꂽZ̃p[eBNƂ̋͏vZ
 * @param[in] gridPos Obhʒu
 * @param[in] i p[eBNCfbNX
 * @param[in] pos0 vZW
 * @param[in] vel0 vZW̑x
 * @param[in] dens0 vZW̖x
 * @param[in] dens p[eBNx
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZ͏
 */
__device__
float3 calExtForceCell(int3 gridPos, uint i, float3 pos0, float3 vel0, float dens0, float* dens, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float h = params.EffectiveRadius;

	float3 frc = make_float3(0.0f);
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);
		for(uint j = startIndex; j < endIndex; ++j){
			if(j != i){
				// ߖTp[eBÑp[^
				float3 pos1 = make_float3(FETCHC(dSortedPos, j));
				float3 vel1 = make_float3(FETCHC(dSortedVel, j));

				float3 rij = pos0-pos1;
				float r = length(rij);

				if(r <= h && r > 0.0001){
					float dens1 = dens[cell.dSortedIndex[j]];

					float3 vij = vel1-vel0;

					float q = h-r;

					// S
					frc += params.Viscosity*params.Mass*(vij/dens1)*params.LWvisc*q;
				}
			}
		}
	}

	return frc;
}

/*!
 * p[eBNɂO͂̌vZ(J[l֐)
 * @param[in] dens p[eBNx
 * @param[out] outFrc p[eBNɂ
 * @param[in] cell p[eBNObhf[^
 */
__global__
void pbfCalExternalForces(float* dens, float4* outFrc, rxParticleCell cell)
{
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	// \[gςݔz񂩂p[eBNf[^擾
	float3 pos0 = make_float3(FETCHC(dSortedPos, index));
	float3 vel0 = make_float3(FETCHC(dSortedVel, index));
	float h = params.EffectiveRadius;

	// p[eBÑ\[gȂzł̃CfbNX
	uint oIdx = cell.dSortedIndex[index];

	float3 frc = make_float3(0.0f);
	float dens0 = dens[oIdx];

	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos0-make_float3(h));
	grid_pos1 = calcGridPos(pos0+make_float3(h));

	// ͂̃Obh܂߂ċߖTTC͍CSvZ
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);

				frc += calExtForceCell(n_grid_pos, index, pos0, vel0, dens0, dens, cell);
			}
		}
	}

	// O(d)
	frc += params.Gravity;

	outFrc[oIdx] = make_float4(frc, 0.0f);
}


/*!
 * ^ꂽZ̃p[eBNƂ̋XP[Ot@N^̕ꍀvZ
 * @param[in] gridPos Obhʒu
 * @param[in] i p[eBNCfbNX
 * @param[in] pos0 vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calScalingFactorCell(int3 gridPos, uint i, float3 pos0, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float h = params.EffectiveRadius;
	float r0 = params.Density;
	float sd = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);
		for(uint j = startIndex; j < endIndex; ++j){
			if(j == i) continue;

			float3 pos1 = make_float3(FETCHC(dSortedPos, j));

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h && r > 0.0){
				float q = h-r;

				// SpikyJ[lňʒuϓvZ
				float3 dp = (params.GWspiky*q*q*rij/r)/r0;

				sd += dot(dp, dp);
			}

		}
	}

	return sd;
}

/*!
 * XP[Ot@N^̌vZ
 * @param[in] ppos p[eBNSW
 * @param[out] pdens p[eBNx
 * @param[out] pscl XP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in] cell p[eBNObhf[^
 */
__global__
void pbfCalScalingFactor(float4* ppos, float* pdens, float* pscl, float eps, rxParticleCell cell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu

	float h = params.EffectiveRadius;
	float r0 = params.Density;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ͂̃Obh܂߂ċߖTTCxvZ
	float dens = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calDensityCellPB(n_grid_pos, index, pos, cell);
			}
		}
	}

	// xS((1))
	float C = dens/r0-1.0;

	// ͂̃Obh܂߂ċߖTTCXP[Ot@N^̕ꍀvZ
	float sd = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				sd += calScalingFactorCell(n_grid_pos, index, pos, cell);
			}
		}
	}

	// p[eBÑ\[gȂzł̃CfbNX
	uint oIdx = cell.dSortedIndex[index];

	// XP[Ot@N^̌vZ((11))
	pscl[oIdx] = -C/(sd+eps);

	// XVꂽx
	pdens[oIdx] = dens;
}


/*!
 * ^ꂽZ̃p[eBNƂ̋XP[Ot@N^̕ꍀvZ
 * @param[in] gridPos Obhʒu
 * @param[in] index p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float3 calPositionCorrectionCell(int3 gridPos, uint i, float3 pos0, float* pscl, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float k = params.AP_K;
	float n = params.AP_N;
	float wq = params.AP_WQ;

	float h = params.EffectiveRadius;
	float r0 = params.Density;
	float3 dp = make_float3(0.0);

	float dt = params.Dt;

	float si = pscl[cell.dSortedIndex[i]];

	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);
		for(uint j = startIndex; j < endIndex; ++j){
			if(j == i) continue;

			float3 pos1 = make_float3(FETCHC(dSortedPos, j));

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h && r > 0.0){
				float scorr = 0.0f;

				if(params.AP){
					float q1 = h*h-r*r;
					float ww = params.Wpoly6*q1*q1*q1/wq;
					scorr = -k*pow(ww, n)*dt*dt;
				}
				float q = h-r;
				float sj = pscl[cell.dSortedIndex[j]];

				// SpikyJ[lňʒuCʂvZ
				dp += (si+sj+scorr)*(params.GWspiky*q*q*rij/r)/r0;
			}

		}
	}

	return dp;
}

/*!
 * XP[Ot@N^̌vZ
 * @param[in] ppos p[eBNSW
 * @param[out] pdens p[eBNx
 * @param[out] pscl XP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in]  cell p[eBNObhf[^
 */
__global__
void pbfPositionCorrection(float4* ppos, float* pscl, float4* pdp, rxParticleCell cell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu

	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ͂̃Obh܂߂ċߖTTCʒuCʂvZ
	float3 dpij = make_float3(0.0f);
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dpij += calPositionCorrectionCell(n_grid_pos, index, pos, pscl, cell);
			}
		}
	}

	// p[eBÑ\[gȂzł̃CfbNX
	uint oIdx = cell.dSortedIndex[index];

	// ʒuC
	pdp[oIdx] = make_float4(dpij, 0.0);
}

/*!
 * p[eBNʒuC
 * @param[inout] pos p[eBNʒu
 * @param[in] pdp ʒuC
 * @param[in] nprts p[eBN
 */
__global__
void pbfCorrectPosition(float4* ppos, float4* pdp, uint nprts)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= nprts) return;

	// ʒuC
	ppos[index] += pdp[index];
}

/*!
 * xϓ̌vZ
 * @param[inout] pos p[eBNʒu
 * @param[in] pdp ʒuC
 * @param[in] nprts p[eBN
 */
__global__
void pbfDensityFluctuation(float* perr, float* pdens, float rest_dens, uint nprts)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= nprts) return;

	// xϓ
	//perr[index] = fabs(pdens[index]-rest_dens)/rest_dens;
	float err = pdens[index]-rest_dens;
	perr[index] = (err >= 0.0f ? err : 0.0f)/rest_dens;
}




/*!
 * ^ꂽZ̃p[eBNƂ̋疧xvZ
 * @param[in] gridPos Obhʒu
 * @param[in] index p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calBoundaryDensityCellPB(int3 gridPos, uint i, float3 pos0, float* dVolB, rxParticleCell bcell)
{
	uint gridHash = calcGridHashB(gridPos, params.GridSizeB);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = bcell.dCellStart[gridHash];

	float h = params.EffectiveRadius;
	float dens = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = bcell.dCellEnd[gridHash];
		for(uint j = startIndex; j < endIndex; ++j){
			//if(j == i) continue;

			float3 pos1 = make_float3(bcell.dSortedPos[j]);
			uint jdx = bcell.dSortedIndex[j];

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h){
				float q = h*h-r*r;
				dens += params.Density*dVolB[jdx]*params.Wpoly6*q*q*q;
			}
		}
	}

	return dens;
}

/*!
 * p[eBNxvZ(J[l֐)
 * @param[out] newDens p[eBNx
 * @param[out] newPres p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
__global__
void pbfCalBoundaryDensity(float* newDens, float4* dPos, float* dVolB, rxParticleCell bcell, uint pnum)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= pnum) return;	
	
	float3 pos = make_float3(dPos[index]);	// p[eBNʒu
	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos1 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);

	// ͂̃Obh܂߂ċߖTTCxvZ
	float dens = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calBoundaryDensityCellPB(n_grid_pos, index, pos, dVolB, bcell);
			}
		}
	}

	// xʂɏ
	newDens[index] += dens;
}




/*!
 * ^ꂽZ̃p[eBNƂ̋XP[Ot@N^̕ꍀvZ
 * @param[in] gridPos Obhʒu
 * @param[in] index p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calBoundaryScalingFactorCell(int3 gridPos, uint i, float3 pos0, float* dVolB, rxParticleCell bcell)
{
	uint gridHash = calcGridHashB(gridPos, params.GridSizeB);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = bcell.dCellStart[gridHash];

	float h = params.EffectiveRadius;
	float r0 = params.Density;
	float sd = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = bcell.dCellEnd[gridHash];
		for(uint j = startIndex; j < endIndex; ++j){
			float3 pos1 = make_float3(bcell.dSortedPos[j]);
			uint jdx = bcell.dSortedIndex[j];

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h && r > 0.0){
				float q = h-r;

				// SpikyJ[lňʒuϓvZ
				float3 dp = (params.Density*dVolB[jdx]/params.Mass)*(params.GWspiky*q*q*rij/r)/r0;

				sd += dot(dp, dp);
			}

		}
	}

	return sd;
}

/*!
 * XP[Ot@N^̌vZ(Ep[eBN܂)
 * @param[in] ppos p[eBNSW
 * @param[out] pdens p[eBNx
 * @param[out] pscl XP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in] cell p[eBNObhf[^
 */
__global__
void pbfCalScalingFactorWithBoundary(float4* ppos, float* pdens, float* pscl, float eps, rxParticleCell cell, 
										float* bvol, rxParticleCell bcell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu

	float h = params.EffectiveRadius;
	float r0 = params.Density;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ̃p[eBNɂ閧x
	float dens = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calDensityCellPB(n_grid_pos, index, pos, cell);
			}
		}
	}

	// p[eBN͂̃Obh
	int3 grid_pos2, grid_pos3;
	grid_pos2 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos3 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);

	// Ep[eBNɂ閧x
	for(int z = grid_pos2.z; z <= grid_pos3.z; ++z){
		for(int y = grid_pos2.y; y <= grid_pos3.y; ++y){
			for(int x = grid_pos2.x; x <= grid_pos3.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calBoundaryDensityCellPB(n_grid_pos, index, pos, bvol, bcell);
			}
		}
	}

	// xS((1))
	float C = dens/r0-1.0;

	// ̃p[eBNɂXP[Ot@N^̕ꍀvZ
	float sd = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				sd += calScalingFactorCell(n_grid_pos, index, pos, cell);
			}
		}
	}

	// Ep[eBNɂXP[Ot@N^̕ꍀvZ
	for(int z = grid_pos2.z; z <= grid_pos3.z; ++z){
		for(int y = grid_pos2.y; y <= grid_pos3.y; ++y){
			for(int x = grid_pos2.x; x <= grid_pos3.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				sd += calBoundaryScalingFactorCell(n_grid_pos, index, pos, bvol, bcell);
			}
		}
	}

	// p[eBÑ\[gȂzł̃CfbNX
	uint oIdx = cell.dSortedIndex[index];

	// XP[Ot@N^̌vZ((11))
	pscl[oIdx] = -C/(sd+eps);

	// XVꂽx
	pdens[oIdx] = dens;
}



/*!
 * XP[Ot@N^̌vZ(Ep[eBN܂)
 * @param[in] ppos p[eBNSW
 * @param[out] pdens p[eBNx
 * @param[out] pscl XP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in] cell p[eBNObhf[^
 */
__global__
void pbfCalBoundaryScalingFactor(float4* ppos, float* pdens, float eps, rxParticleCell cell, 
									float* bvol, float* bscl, rxParticleCell bcell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= bcell.uNumParticles) return;	
	
	float3 pos = make_float3(bcell.dSortedPos[index]);	// p[eBNʒu

	float h = params.EffectiveRadius;
	float r0 = params.Density;

	// p[eBN͂̃Obh(̃p[eBNp)
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ̃p[eBNɂ閧x
	float dens = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calDensityCellPB(n_grid_pos, index, pos, cell);
			}
		}
	}

	// p[eBN͂̃Obh(Ep[eBNp)
	int3 grid_pos2, grid_pos3;
	grid_pos2 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos3 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);

	// Ep[eBNɂ閧x
	for(int z = grid_pos2.z; z <= grid_pos3.z; ++z){
		for(int y = grid_pos2.y; y <= grid_pos3.y; ++y){
			for(int x = grid_pos2.x; x <= grid_pos3.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dens += calBoundaryDensityCellPB(n_grid_pos, index, pos, bvol, bcell);
			}
		}
	}

	// xS((1))
	float C = dens/r0-1.0;

	// ̃p[eBNɂXP[Ot@N^̕ꍀvZ
	float sd = 0.0f;
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				sd += calScalingFactorCell(n_grid_pos, index, pos, cell);
			}
		}
	}

	// Ep[eBNɂXP[Ot@N^̕ꍀvZ
	for(int z = grid_pos2.z; z <= grid_pos3.z; ++z){
		for(int y = grid_pos2.y; y <= grid_pos3.y; ++y){
			for(int x = grid_pos2.x; x <= grid_pos3.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				sd += calBoundaryScalingFactorCell(n_grid_pos, index, pos, bvol, bcell);
			}
		}
	}

	// p[eBÑ\[gȂzł̃CfbNX
	uint oIdx = bcell.dSortedIndex[index];

	// XP[Ot@N^̌vZ((11))
	bscl[oIdx] = -C/(sd+eps);
}



/*!
 * ^ꂽZ̃p[eBNƂ̋XP[Ot@N^̕ꍀvZ
 * @param[in] gridPos Obhʒu
 * @param[in] index p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float3 calBoundaryPositionCorrectionCell(int3 gridPos, uint i, float3 pos0, float si, float* bscl, float* bvol, rxParticleCell bcell)
{
	uint gridHash = calcGridHashB(gridPos, params.GridSizeB);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = bcell.dCellStart[gridHash];

	float k = params.AP_K;
	float n = params.AP_N;
	float wq = params.AP_WQ;

	float h = params.EffectiveRadius;
	float r0 = params.Density;
	float3 dp = make_float3(0.0);

	float dt = params.Dt;

	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = bcell.dCellEnd[gridHash];
		for(uint j = startIndex; j < endIndex; ++j){
			float3 pos1 = make_float3(bcell.dSortedPos[j]);
			uint jdx = bcell.dSortedIndex[j];

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h && r > 0.0){
				float scorr = 0.0f;

				if(params.AP){
					float q1 = h*h-r*r;
					float ww = (params.Density*bvol[jdx]/params.Mass)*params.Wpoly6*q1*q1*q1/wq;
					scorr = -k*pow(ww, n)*dt*dt;
				}
				float q = h-r;
				float sj = bscl[jdx];

				// SpikyJ[lňʒuCʂvZ
				dp += (si+sj+scorr)*(params.GWspiky*q*q*rij/r)/r0;
			}

		}
	}

	return dp;
}

/*!
 * XP[Ot@N^̌vZ
 * @param[in] ppos p[eBNSW
 * @param[out] pdens p[eBNx
 * @param[out] pscl XP[Ot@N^
 * @param[in] eps ɘaW
 * @param[in]  cell p[eBNObhf[^
 */
__global__
void pbfPositionCorrectionWithBoundary(float4* ppos, float* pscl, float4* pdp, rxParticleCell cell, 
										  float* bvol, float* bscl, rxParticleCell bcell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu

	float h = params.EffectiveRadius;

	float si = pscl[cell.dSortedIndex[index]];


	// p[eBN͂̃Obh(̃p[eBNp)
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ̃p[eBNɂʒuCʂvZ
	float3 dpij = make_float3(0.0f);
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dpij += calPositionCorrectionCell(n_grid_pos, index, pos, pscl, cell);
			}
		}
	}

	// p[eBN͂̃Obh(Ep[eBNp)
	int3 grid_pos2, grid_pos3;
	grid_pos2 = calcGridPosB(pos-make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);
	grid_pos3 = calcGridPosB(pos+make_float3(h), params.WorldOriginB, params.CellWidthB, params.GridSizeB);

	// Ep[eBNɂʒuCʂvZ
	for(int z = grid_pos2.z; z <= grid_pos3.z; ++z){
		for(int y = grid_pos2.y; y <= grid_pos3.y; ++y){
			for(int x = grid_pos2.x; x <= grid_pos3.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				dpij += calBoundaryPositionCorrectionCell(n_grid_pos, index, pos, si, bscl, bvol, bcell);
			}
		}
	}

	// p[eBÑ\[gȂzł̃CfbNX
	uint oIdx = cell.dSortedIndex[index];

	// ʒuC
	pdp[oIdx] = make_float4(dpij, 0.0);
}


__device__
void calCollisionSolidPB(float3 &pos, float3 &vel, float dt)
{
	float d;
	float3 n;
	float3 cp;

	// {bNX`̃IuWFNgƂ̏Փ
#if MAX_BOX_NUM
	for(int i = 0; i < params.BoxNum; ++i){
		if(params.BoxFlg[i] == 0) continue;
		
		collisionPointBox(pos, params.BoxCen[i], params.BoxExt[i], params.BoxRot[i], params.BoxInvRot[i], cp, d, n);

		if(d < 0.0){
			float res = params.Restitution;
			res = (res > 0) ? (res*fabs(d)/(dt*length(vel))) : 0.0f;
			vel -= (1+res)*n*dot(n, vel);
			pos = cp;
		}
	}
#endif

	// `̃IuWFNgƂ̏Փ
#if MAX_SPHERE_NUM
	for(int i = 0; i < params.SphereNum; ++i){
		if(params.SphereFlg[i] == 0) continue;

		collisionPointSphere(pos, params.SphereCen[i], params.SphereRad[i], cp, d, n);

		if(d < 0.0){
			float res = params.Restitution;
			res = (res > 0) ? (res*fabs(d)/(dt*length(vel))) : 0.0f;
			vel -= (1+res)*n*dot(n, vel);
			pos = cp;
		}
	}
#endif

	// ͂̋EƂ̏Փ˔
	float3 l0 = params.Boundary[0];
	float3 l1 = params.Boundary[1];
	collisionPointAABB(pos, 0.5*(l1+l0), 0.5*(l1-l0), cp, d, n);

	if(d < 0.0){
		float res = params.Restitution;
		res = (res > 0) ? (res*fabs(d)/(dt*length(vel))) : 0.0f;
		vel -= (1+res)*n*dot(n, vel);
		pos = cp;
	}
}

__device__
inline bool calCollisionPolygonPB(float3 &pos0, float3 &pos1, float3 &vel, float3 v0, float3 v1, float3 v2, float dt)
{
	float3 cp, n;
	if(intersectSegmentTriangle(pos0, pos1, v0, v1, v2, cp, n, params.ParticleRadius) == 1){
		float d = length(pos1-cp);
		n = normalize(n);

		//float res = params.Restitution;
		//res = (res > 0) ? (res*fabs(d)/(dt*length(vel))) : 0.0f;
		//float3 vr = -(1+res)*n*dot(n, vel);

		float3 v = pos1-pos0;
		float l = length(v);
		v /= l;
		float3 vd = v*(l-d);
		float3 vr = vd-2*dot(n, vd)*n;

		pos1 = cp+vr*0.7;

		//vel += vr;//+params.PolyVel[0];
		//vel.x = 1.0;

		return true;
	}
	return false;
}




/*!
 * p[eBNʒuCx̍XV
 * @param[inout] ppos p[eBNʒu
 * @param[inout] pvel p[eBNx
 * @param[in] pfrc p[eBNɂ
 * @param[in] dens p[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
__global__
void pbfIntegrate(float4* ppos, float4* pvel, float4* pacc, 
					 float4* new_ppos, float4* new_pvel, float dt, uint nprts)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= nprts) return;

	float3 x = make_float3(ppos[index]);
	float3 v = make_float3(pvel[index]);
	float3 a = make_float3(pacc[index]);
	//float3 v_old = v;

	// XVʒuCx̍XV
	v += dt*a;
	x += dt*v;

	// ő́EEƂ̏Փ
	calCollisionSolidPB(x, v, dt);

	// ʒuƑx̍XV
	new_ppos[index] = make_float4(x);
	new_pvel[index] = make_float4(v);
}



/*!
 * p[eBNʒuCx̍XV(Leap-Frog)
 * @param[inout] ppos p[eBNʒu
 * @param[inout] pvel p[eBNx
 * @param[in] pfrc p[eBNɂ
 * @param[in] dens p[eBNx
 * @param[in] vrts
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
__global__
void pbfIntegrateWithPolygon(float4* ppos, float4* pvel, float4* pacc, 
								float4* new_ppos, float4* new_pvel, 
								float3* vrts, int3* tris, int tri_num, float dt, rxParticleCell cell)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;

	float3 x = make_float3(ppos[index]);
	float3 v = make_float3(pvel[index]);
	float3 a = make_float3(pacc[index]);
	//float3 v_old = v;
	float3 x_old = x;

	// XVʒuCx̍XV
	v += dt*a;
	x += dt*v;

	// |SIuWFNgƂ̏Փ
	int3 gridPos[2];
	gridPos[0] = calcGridPos(x_old);	// ʒuXVÕp[eBNObh
	gridPos[1] = calcGridPos(x);		// ʒuXṼp[eBNObh
	for(int i = 0; i < 2; ++i){
		uint grid_hash = calcGridHash(gridPos[i]);
		uint start_index = cell.dPolyCellStart[grid_hash];
		if(start_index != 0xffffffff){	// ZłȂ̃`FbN

			uint end_index = cell.dPolyCellEnd[grid_hash];
			for(uint j = start_index; j < end_index; ++j){
				uint pidx = cell.dSortedPolyIdx[j];

				int3 idx = tris[pidx];
				if(calCollisionPolygonPB(x_old, x, v, vrts[idx.x], vrts[idx.y], vrts[idx.z], dt)){
				}
			}
		}
	}

	// ő́EEƂ̏Փ
	calCollisionSolidPB(x, v, dt);

	// ʒuƑx̍XV
	new_ppos[index] = make_float4(x);
	new_pvel[index] = make_float4(v);
}



/*!
 * p[eBNʒuCx̍XV
 * @param[inout] ppos p[eBNʒu
 * @param[inout] pvel p[eBNx
 * @param[in] pfrc p[eBNɂ
 * @param[in] dens p[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
__global__
void pbfIntegrate2(float4* ppos, float4* pvel, float4* pacc, 
					  float4* new_ppos, float4* new_pvel, float dt, uint nprts)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= nprts) return;

	float3 x = make_float3(new_ppos[index]);
	float3 v = make_float3(new_pvel[index]);

	// ő́EEƂ̏Փ
	calCollisionSolidPB(x, v, dt);

	// ʒuƑx̍XV
	new_ppos[index] = make_float4(x);
}



/*!
 * p[eBNʒuCx̍XV(Leap-Frog)
 * @param[inout] ppos p[eBNʒu
 * @param[inout] pvel p[eBNx
 * @param[in] pfrc p[eBNɂ
 * @param[in] dens p[eBNx
 * @param[in] vrts
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
__global__
void pbfIntegrateWithPolygon2(float4* ppos, float4* pvel, float4* pacc, 
							  float4* new_ppos, float4* new_pvel, 
							  float3* vrts, int3* tris, int tri_num, float dt, rxParticleCell cell)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;

	float3 x = make_float3(new_ppos[index]);
	float3 x_old = make_float3(ppos[index]);
	float3 v = make_float3(new_pvel[index]);

	// |SIuWFNgƂ̏Փ
	int3 gridPos[2];
	gridPos[0] = calcGridPos(x_old);	// ʒuXVÕp[eBNObh
	gridPos[1] = calcGridPos(x);		// ʒuXṼp[eBNObh
	for(int i = 0; i < 2; ++i){
		uint grid_hash = calcGridHash(gridPos[i]);
		uint start_index = cell.dPolyCellStart[grid_hash];
		if(start_index != 0xffffffff){	// ZłȂ̃`FbN

			uint end_index = cell.dPolyCellEnd[grid_hash];
			for(uint j = start_index; j < end_index; ++j){
				uint pidx = cell.dSortedPolyIdx[j];

				int3 idx = tris[pidx];
				if(calCollisionPolygonPB(x_old, x, v, vrts[idx.x], vrts[idx.y], vrts[idx.z], dt)){
				}
			}
		}
	}

	// ő́EEƂ̏Փ
	calCollisionSolidPB(x, v, dt);

	// ʒuƑx̍XV
	new_ppos[index] = make_float4(x);
	new_pvel[index] = make_float4(v);
}


/*!
 * p[eBNʒuCx̍XV
 * @param[in] ppos XVꂽp[eBNʒu
 * @param[inout] new_ppos Xebvŏ̃p[eBNʒu/Vp[eBNx
 * @param[out] new_pvel Vp[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
__global__
void pbfUpdatePosition(float4* ppos, float4* new_ppos, float4* new_pvel, float dt, uint nprts)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= nprts) return;

	float3 x0 = make_float3(new_ppos[index]);
	float3 x1 = make_float3(ppos[index]);
	float3 v = (x1-x0)/dt;

	// ʒuƑx̍XV
	new_pvel[index] = make_float4(v);
	new_ppos[index] = make_float4(x1);
}

/*!
 * p[eBNx̍XV
 * @param[in] ppos XVꂽp[eBNʒu
 * @param[in] new_ppos Xebvŏ̃p[eBNʒu/Vp[eBNx
 * @param[out] new_pvel Vp[eBNx
 * @param[in] dt ԃXebv
 * @param[in] nprts p[eBN
 */
__global__
void pbfUpdateVelocity(float4* ppos, float4* new_ppos, float4* new_pvel, float dt, uint nprts)
{
	uint index = __umul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= nprts) return;

	float3 x0 = make_float3(new_ppos[index]);
	float3 x1 = make_float3(ppos[index]);
	float3 v = (x1-x0)/dt;

	// ʒuƑx̍XV
	new_pvel[index] = make_float4(v);
}



/*!
 * ^ꂽZ̃p[eBNƂ̋疧xvZ
 * @param[in] gridPos Obhʒu
 * @param[in] index p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float3 calXsphViscosityCell(int3 gridPos, uint i, float3 pos0, float3 vel0, float4* pvel, float* dens, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float h = params.EffectiveRadius;
	float3 v = make_float3(0.0);
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);
		for(uint j = startIndex; j < endIndex; ++j){
			//if(j == i) continue;

			float3 pos1 = make_float3(FETCHC(dSortedPos, j));

			float3 rij = pos0-pos1;
			float r = length(rij);

			if(r <= h){
				float3 vel1 = make_float3(pvel[cell.dSortedIndex[j]]);
				float3 rho1 = make_float3(dens[cell.dSortedIndex[j]]);

				float q = h*h-r*r;
				v += (params.Mass/rho1)*(vel1-vel0)*params.Wpoly6*q*q*q;
			}
		}
	}

	return v;
}

/*!
 * p[eBNxvZ(J[l֐)
 * @param[out] newDens p[eBNx
 * @param[out] newPres p[eBN
 * @param[in]  cell p[eBNObhf[^
 */
__global__
void xsphVisocosity(float4* ppos, float4* pvel, float4* new_pvel, float* dens, float c, rxParticleCell cell)
{
	// p[eBNCfbNX
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos0 = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	float3 vel0 = make_float3(pvel[cell.dSortedIndex[index]]);	// p[eBNx
	//int3 grid_pos = calcGridPos(pos0);	// p[eBNObhʒu
	float h = params.EffectiveRadius;

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos0-make_float3(h));
	grid_pos1 = calcGridPos(pos0+make_float3(h));

	// ͂̃Obh܂߂ċߖTTCxvZ
	float3 v = make_float3(0.0);
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				v += calXsphViscosityCell(n_grid_pos, index, pos0, vel0, pvel, dens, cell);
			}
		}
	}

	// xƈ͒lʂɏ
	uint oIdx = cell.dSortedIndex[index];
	new_pvel[oIdx] = make_float4(vel0+c*v);
	//new_pvel[oIdx] = make_float4(vel0);
}




/*!
 * ^ꂽZ̃p[eBNƂ̋疧xvZ
 * @param[in] gridPos Obhʒu
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float calDensityCellGPB(int3 gridPos, float3 pos0, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float h = params.EffectiveRadius;
	float d = 0.0f;
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);

		for(uint j = startIndex; j < endIndex; ++j){
			//if(j != index){
				float3 pos1 = make_float3(FETCHC(dSortedPos, j));

				float3 rij = pos0-pos1;
				float r = length(rij);

				if(r <= h){
					float q = h*h-r*r;

					d += params.Mass*params.Wpoly6*q*q*q;
				}

			//}
		}
	}

	return d;
}

/*!
 * Obhł̖xvZ
 * @param[out] GridD Obhx
 * @param[in] cell p[eBNObhf[^
 * @param[in] gnum Obh
 * @param[in] gmin ObhŏW
 * @param[in] glen Obh
 */
__global__
void pbfCalDensityInGrid(float* GridD, rxParticleCell cell, 
					uint3 gnum, float3 gmin, float3 glen)
{
	uint blockId = __mul24(blockIdx.y, gridDim.x)+blockIdx.x;
	uint i = __mul24(blockId, blockDim.x)+threadIdx.x;

	uint3 gridPos = calcGridPosU(i, gnum);

	if(gridPos.x < gnum.x && gridPos.y < gnum.y && gridPos.z < gnum.z){
		float3 gpos;
		gpos.x = gmin.x+(gridPos.x)*glen.x;
		gpos.y = gmin.y+(gridPos.y)*glen.y;
		gpos.z = gmin.z+(gridPos.z)*glen.z;

		float d = 0.0f;

		int3 pgpos = calcGridPos(gpos);

		float h = params.EffectiveRadius;
		int3 grid_pos0, grid_pos1;
		grid_pos0 = calcGridPos(gpos-make_float3(h));
		grid_pos1 = calcGridPos(gpos+make_float3(h));

		for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
			for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
				for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
					int3 neighbourPos = make_int3(x, y, z);

					d += calDensityCellGPB(neighbourPos, gpos, cell);
				}
			}
		}

		GridD[gridPos.x+gridPos.y*gnum.x+gridPos.z*gnum.x*gnum.y] = d;
	}

}

/*!
 * ^ꂽZ̃p[eBNƂ̋@vZ
 * @param[in] gridPos Obhʒu
 * @param[in] i p[eBNCfbNX
 * @param[in] pos vZW
 * @param[in] cell p[eBNObhf[^
 * @return Z̃p[eBNvZxl
 */
__device__
float3 calNormalCellPB(int3 gridPos, uint i, float3 pos0, float* dens, rxParticleCell cell)
{
	uint gridHash = calcGridHash(gridPos);

	// Z̃p[eBÑX^[gCfbNX
	uint startIndex = FETCHC(dCellStart, gridHash);

	float h = params.EffectiveRadius;
	float3 nrm = make_float3(0.0f);
	if(startIndex != 0xffffffff){	// ZłȂ̃`FbN
		// Z̃p[eBNŔ
		uint endIndex = FETCHC(dCellEnd, gridHash);

		for(uint j = startIndex; j < endIndex; ++j){
			if(j != i){
				float3 pos1 = make_float3(FETCHC(dSortedPos, j));

				float3 rij = pos0-pos1;
				float r = length(rij);

				if(r <= h && r > 0.0001){
					float d1 = dens[cell.dSortedIndex[j]];
					float q = h*h-r*r;

					nrm += (params.Mass/d1)*params.GWpoly6*q*q*rij;
				}

			}
		}
	}

	return nrm;
}


/*!
 * p[eBN@vZ(J[l֐)
 * @param[out] newNrms p[eBN@
 * @param[in] dens p[eBNx
 * @param[in] cell p[eBNObhf[^
 */
__global__
void pbfCalNormal(float4* newNrms, float* dens, rxParticleCell cell)
{
	uint index = __mul24(blockIdx.x,blockDim.x)+threadIdx.x;
	if(index >= cell.uNumParticles) return;	
	
	float3 pos = make_float3(FETCHC(dSortedPos, index));	// p[eBNʒu
	float h = params.EffectiveRadius;
	//int3 grid_pos = calcGridPos(pos);	// p[eBNObhʒu

	// p[eBN͂̃Obh
	int3 grid_pos0, grid_pos1;
	grid_pos0 = calcGridPos(pos-make_float3(h));
	grid_pos1 = calcGridPos(pos+make_float3(h));

	// ͂̃Obh܂߂ċߖTTCxvZ
	float3 nrm = make_float3(0.0f);
	for(int z = grid_pos0.z; z <= grid_pos1.z; ++z){
		for(int y = grid_pos0.y; y <= grid_pos1.y; ++y){
			for(int x = grid_pos0.x; x <= grid_pos1.x; ++x){
				int3 n_grid_pos = make_int3(x, y, z);
				nrm += calNormalCellPB(n_grid_pos, index, pos, dens, cell);
			}
		}
	}

	float l = length(nrm);
	if(l > 0){
		nrm /= l;
	}

	uint oIdx = cell.dSortedIndex[index];
	newNrms[oIdx] = make_float4(nrm, 0.0f);
}





#endif // #ifndef _RX_PVSPH_KERNEL_CU_



